"use strict";

var _lodash = require("lodash");

function _toConsumableArray(arr) { return _arrayWithoutHoles(arr) || _iterableToArray(arr) || _nonIterableSpread(); }

function _nonIterableSpread() { throw new TypeError("Invalid attempt to spread non-iterable instance"); }

function _iterableToArray(iter) { if (Symbol.iterator in Object(iter) || Object.prototype.toString.call(iter) === "[object Arguments]") return Array.from(iter); }

function _arrayWithoutHoles(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = new Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } }

var joinArrays = require('./join-arrays');

var joinArraysSmart = require('./join-arrays-smart');

function merge() {
  for (var _len = arguments.length, sources = new Array(_len), _key = 0; _key < _len; _key++) {
    sources[_key] = arguments[_key];
  }

  // This supports
  // merge([<object>] | ...<object>)
  // merge({ customizeArray: <fn>, customizeObject: <fn>})([<object>] | ...<object>)
  // where fn = (a, b, key)
  if (sources.length === 1) {
    if (Array.isArray(sources[0])) {
      return _lodash.mergeWith.apply(void 0, [{}].concat(_toConsumableArray(sources[0]), [joinArrays(sources[0])]));
    }

    if (sources[0].customizeArray || sources[0].customizeObject) {
      return function () {
        for (var _len2 = arguments.length, structures = new Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
          structures[_key2] = arguments[_key2];
        }

        if (Array.isArray(structures[0])) {
          return _lodash.mergeWith.apply(void 0, [{}].concat(_toConsumableArray(structures[0]), [joinArrays(sources[0])]));
        }

        return _lodash.mergeWith.apply(void 0, [{}].concat(structures, [joinArrays(sources[0])]));
      };
    }

    return sources[0];
  }

  return _lodash.mergeWith.apply(void 0, [{}].concat(sources, [joinArrays()]));
} // rules: { <field>: <'append'|'prepend'|'replace'> }
// All default to append but you can override here


var mergeStrategy = function mergeStrategy() {
  var rules = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
  return merge({
    customizeArray: _customizeArray(rules),
    customizeObject: customizeObject(rules)
  });
};

var mergeSmartStrategy = function mergeSmartStrategy() {
  var rules = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
  var plugins = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : [];
  return merge({
    customizeArray: function customizeArray(a, b, key) {
      var topKey = key.split('.').slice(-1)[0];

      if (isRule(topKey)) {
        var _uniteRules = function _uniteRules(aRule, bRule) {
          return joinArraysSmart.uniteRules(aRule, bRule, rules[key]);
        };

        switch (rules[key]) {
          case 'prepend':
            return [].concat(_toConsumableArray((0, _lodash.differenceWith)(b, a, _uniteRules)), _toConsumableArray(a));

          case 'replace':
            return b;

          default:
            // append
            return (0, _lodash.unionWith)(a, b, _uniteRules);
        }
      } else if (topKey === 'plugins') {
        switch (rules[key]) {
          case 'prepend':
            return [].concat(_toConsumableArray((0, _lodash.differenceWith)(b, a, joinArraysSmart.unitePlugins.bind(null, plugins))), _toConsumableArray(a));

          case 'replace':
            return b;

          default:
            // append
            return (0, _lodash.unionWith)(a, b, joinArraysSmart.unitePlugins.bind(null, plugins));
        }
      }

      return _customizeArray(rules)(a, b, key);
    },
    customizeObject: customizeObject(rules)
  });
};

function _customizeArray(rules) {
  return function (a, b, key) {
    switch (rules[key]) {
      case 'prepend':
        return [].concat(_toConsumableArray(b), _toConsumableArray(a));

      case 'replace':
        return b;

      default:
        // append
        return false;
    }
  };
}

function customizeObject(rules) {
  return function (a, b, key) {
    switch (rules[key]) {
      case 'prepend':
        return (0, _lodash.mergeWith)({}, b, a, joinArrays());

      case 'replace':
        return b;

      default:
        // append
        return false;
    }
  };
}

function isRule(key) {
  return ['preLoaders', 'loaders', 'postLoaders', 'rules'].indexOf(key) >= 0;
}

module.exports = merge;
module.exports.strategy = mergeStrategy;
module.exports.smart = mergeSmartStrategy();
module.exports.smartStrategy = mergeSmartStrategy;