"use strict";

var _lodash = require("lodash");

function _slicedToArray(arr, i) { return _arrayWithHoles(arr) || _iterableToArrayLimit(arr, i) || _nonIterableRest(); }

function _nonIterableRest() { throw new TypeError("Invalid attempt to destructure non-iterable instance"); }

function _iterableToArrayLimit(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"] != null) _i["return"](); } finally { if (_d) throw _e; } } return _arr; }

function _arrayWithHoles(arr) { if (Array.isArray(arr)) return arr; }

function _toConsumableArray(arr) { return _arrayWithoutHoles(arr) || _iterableToArray(arr) || _nonIterableSpread(); }

function _nonIterableSpread() { throw new TypeError("Invalid attempt to spread non-iterable instance"); }

function _iterableToArray(iter) { if (Symbol.iterator in Object(iter) || Object.prototype.toString.call(iter) === "[object Arguments]") return Array.from(iter); }

function _arrayWithoutHoles(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = new Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } }

var joinArrays = require('./join-arrays');

var isArray = Array.isArray;

function uniteRules(newRule, rule, strategy) {
  if (String(rule.test) !== String(newRule.test) || newRule.enforce && rule.enforce !== newRule.enforce || newRule.include && !isSameValue(rule.include, newRule.include) || newRule.exclude && !isSameValue(rule.exclude, newRule.exclude)) {
    return false;
  } else if (!rule.test && !rule.include && !rule.exclude && (rule.loader && rule.loader.split('?')[0]) !== (newRule.loader && newRule.loader.split('?')[0])) {
    // Don't merge the rule if there isn't any identifying fields and the loaders don't match
    return false;
  } // webpack 2 nested rules support


  if (rule.rules) {
    var _uniteRules = function _uniteRules(b, a) {
      return uniteRules(b, a, strategy);
    };

    switch (strategy) {
      case 'prepend':
        rule.rules = [].concat(_toConsumableArray((0, _lodash.differenceWith)(newRule.rules, rule.rules, _uniteRules)), _toConsumableArray(rule.rules));
        break;

      default:
        rule.rules = (0, _lodash.unionWith)(rule.rules, newRule.rules, uniteRules);
        break;
    }
  } // newRule.loader should always override


  if (newRule.loader) {
    var optionsKey = newRule.options ? 'options' : newRule.query && 'query';
    delete rule.use;
    delete rule.loaders;
    rule.loader = newRule.loader;

    if (optionsKey) {
      rule[optionsKey] = newRule[optionsKey];
    }
  } else if ((rule.use || rule.loaders || rule.loader) && (newRule.use || newRule.loaders)) {
    var expandEntry = function expandEntry(loader) {
      return typeof loader === 'string' ? {
        loader: loader
      } : loader;
    }; // this is only here to avoid breaking existing tests


    var unwrapEntry = function unwrapEntry(entry) {
      return !entry.options && !entry.query ? entry.loader : entry;
    };

    var entries;

    if (rule.loader) {
      var _optionsKey = rule.options ? 'options' : rule.query && 'query';

      entries = [{
        loader: rule.loader
      }];

      if (_optionsKey) {
        entries[0][_optionsKey] = rule[_optionsKey];
      }

      delete rule.loader;

      if (_optionsKey) {
        delete rule[_optionsKey];
      }
    } else {
      entries = [].concat(rule.use || rule.loaders).map(expandEntry);
    }

    var newEntries = [].concat(newRule.use || newRule.loaders).map(expandEntry);
    var loadersKey = rule.use || newRule.use ? 'use' : 'loaders';

    switch (strategy) {
      case 'prepend':
        rule[loadersKey] = [].concat(_toConsumableArray((0, _lodash.differenceWith)(newEntries, entries, uniteEntries)), _toConsumableArray(entries)).map(unwrapEntry);
        break;

      default:
        rule[loadersKey] = (0, _lodash.unionWith)(entries, newEntries, uniteEntries).map(unwrapEntry);
        break;
    }
  }

  if (newRule.include) {
    rule.include = newRule.include;
  }

  if (newRule.exclude) {
    rule.exclude = newRule.exclude;
  }

  return true;
}
/**
 * Check equality of two values using lodash's isEqual
 * Arrays need to be sorted for equality checking
 * but clone them first so as not to disrupt the sort order in tests
 */


function isSameValue(a, b) {
  var _map = [a, b].map(function (value) {
    return isArray(value) ? _toConsumableArray(value).sort() : value;
  }),
      _map2 = _slicedToArray(_map, 2),
      propA = _map2[0],
      propB = _map2[1];

  return (0, _lodash.isEqual)(propA, propB);
}

function uniteEntries(newEntry, entry) {
  var loaderNameRe = /^([^?]+)/ig;

  var _entry$loader$match = entry.loader.match(loaderNameRe),
      _entry$loader$match2 = _slicedToArray(_entry$loader$match, 1),
      loaderName = _entry$loader$match2[0];

  var _newEntry$loader$matc = newEntry.loader.match(loaderNameRe),
      _newEntry$loader$matc2 = _slicedToArray(_newEntry$loader$matc, 1),
      newLoaderName = _newEntry$loader$matc2[0];

  if (loaderName !== newLoaderName) {
    return false;
  } // Replace query values with newer ones


  (0, _lodash.mergeWith)(entry, newEntry);
  return true;
}

function unitePlugins(newPlugin, plugin) {
  if (!(newPlugin instanceof plugin.constructor)) return false; // functions shouldn't be merged here

  (0, _lodash.mergeWith)(plugin, newPlugin, function (a, b, k) {
    return (0, _lodash.isFunction)(a) ? a : joinArrays()(a, b, k);
  });
  return true;
}

exports.uniteRules = uniteRules;
exports.uniteEntries = uniteEntries;
exports.unitePlugins = unitePlugins;